<?php

namespace DigitalPoint\Cloudflare\Admin\Controller;

use XF\Mvc\ParameterBag;
use XF\Admin\Controller\AbstractController;

use function count, is_array;

class Cloudflare extends AbstractController
{
	protected function preDispatchController($action, ParameterBag $params)
	{
		if ($action == 'Analytics')
		{
			$this->assertAdminPermission('viewStatistics');
		}
		else
		{
			$this->assertAdminPermission('cloudflare');
		}

		$options = $this->options();

		if ((empty($options->cloudflareAuth['email']) || empty($options->cloudflareAuth['api_key'])) && empty($options->cloudflareAuth['token']))
		{
			throw $this->exception($this->error(\XF::phrase('cloudflare_api_credentials_missing', ['url' => $this->buildLink('options/groups/cloudflare/') . '#cloudflareAuth'])));
		}
	}

	protected function assertCanZip()
	{
		if (!class_exists('ZipArchive'))
		{
			throw $this->exception($this->error(\XF::phrase('backup_requires_ziparchive')));
		}
	}

	protected function assertHasToken()
	{
		if (empty($this->options()->cloudflareAuth['token']))
		{
			throw $this->exception($this->error(\XF::phrase('cloudflare_auth_token_required', ['url' => $this->buildLink('options/groups/cloudflare/') . '#cloudflareAuth'])));
		}
	}

	public function actionIndex(ParameterBag $params)
	{
		$viewParams = $this->getCloudflareRepo()->organizeSettings();
		return $this->view('DigitalPoint\Cloudflare:Settings', 'cloudflare_settings', $viewParams);
	}

	public function actionEasy()
	{
		if ($this->isPost())
		{
			$this->getCloudflareRepo()->setEasyMode();
			return $this->redirect($this->buildLink('cloudflare'));
		}
		return $this->view('DigitalPoint\Cloudflare:Easy', 'cloudflare_easy_config_confirm', []);
	}

	public function actionAnalytics()
	{
		$this->assertPostOnly();

		$range = $this->filter('range', 'str');

		switch ($range) {
			case 'week':
				$days = 7;
				break;
			case 'month':
				$days = 30;
				break;
			case 'year':
				$days = 364;
				break;
			default:
				$days = 1;
		}

		/** @var \DigitalPoint\Cloudflare\Repository\Cloudflare $cloudflareRepo */
		$cloudflareRepo = $this->getCloudflareRepo();

		$promises = ['stats' => $cloudflareRepo->getGraphQLZoneAnalytics($days)];
		$promises = $cloudflareRepo->resolvePromises($promises);
		$stats = $cloudflareRepo->prepareGraphQLZoneAnalytics($promises['stats']);

		$reply = $this->view('DigitalPoint\Cloudflare:Analytics');
		$reply->setJsonParams([
			'stats' => $stats,
		]);
		return $reply;
	}

	public function actionAnalyticsDmarc()
	{
		$this->assertPostOnly();

		$range = $this->filter('range', 'str');

		switch ($range) {
			case 'week':
				$days = 7;
				break;
			case 'month':
				$days = 30;
				break;
			case 'year':
				$days = 364;
				break;
			default:
				$days = 1;
		}

		/** @var \DigitalPoint\Cloudflare\Repository\Cloudflare $cloudflareRepo */
		$cloudflareRepo = $this->getCloudflareRepo();

		$promises = ['stats' => $cloudflareRepo->getGraphQLZoneAnalyticsDmarc($days)];
		$promises = $cloudflareRepo->resolvePromises($promises);

		$stats = $cloudflareRepo->prepareGraphQLZoneAnalyticsDmarc($promises['stats']);

		$reply = $this->view('DigitalPoint\Cloudflare:Analytics\Dmarc');
		$reply->setJsonParams([
			'stats' => $stats,
		]);
		return $reply;
	}

	public function actionToggle()
	{
		$cloudflareRepo = $this->getCloudflareRepo();

		$cloudflareSettings = $cloudflareRepo->organizeSettings(null, false)['settings'];

		$settingsToChange = [];

		foreach ($cloudflareSettings as $setting)
		{
			if (!empty($setting['options']['editable']))
			{
				if (!empty($setting['defaults']['type']) && $setting['defaults']['type'] == 'checkbox')
				{
					$newArray = $this->filter($setting['id'], 'array-str');
					$newValue = [];

					foreach ($setting['options']['value'] as $key => $value)
					{
						$newValue[$key] = empty($newArray[$key]) ? 'off' : $newArray[$key];
						$this->convertBoolean($newValue[$key]);
					}
					if ($newValue != $setting['options']['value'])
					{
						$settingsToChange[$setting['id']] = ['id' => $setting['id'], 'value' => $newValue];
					}
				}
				// For bot fight mode
				elseif(!empty($setting['defaults']['type']) && $setting['defaults']['type'] == 'array' && !is_array($setting['options']['value']))
				{
					$newValue = $this->filter($setting['id'], $setting['defaults']['data_type'][$setting['defaults']['value_key']]);

					$result = $cloudflareRepo->getEndpointResultsByKey($setting['defaults']['override_endpoint'])['result'];

					if ($newValue != $result[$setting['defaults']['value_key']])
					{
						$result[$setting['defaults']['value_key']] = $newValue;
						$settingsToChange[$setting['id']] = $result;
					}
				}
				elseif(!empty($setting['defaults']['type']) && $setting['defaults']['type'] == 'array')
				{
					$newArray = $this->filter($setting['id'], 'array-array-str');
					$newValue = [];

					foreach ($setting['options']['value'] as $key => $value)
					{
						foreach ($value as $subKey => $subValue)
						{
							$newValue[$key][$subKey] = empty($newArray[$key][$subKey]) ? '' : $newArray[$key][$subKey];

							if (!empty($setting['defaults']['data_type'][$key][$subKey]))
							{
								if ($setting['defaults']['data_type'][$key][$subKey] == 'bool')
								{
									$this->convertBoolean($newValue[$key][$subKey], true);
								}
								elseif($setting['defaults']['data_type'][$key][$subKey] == 'int')
								{
									$newValue[$key][$subKey] = intval($newValue[$key][$subKey]);
								}
							}

						}
					}

					if ($newValue != $setting['options']['value'])
					{
						$settingsToChange[$setting['id']] = ['id' => $setting['id'], 'value' => $newValue];
					}

				}
				else
				{
					$newValue = $this->filter($setting['id'], 'str');

					// Need to watch this in case there's a Cloudflare setting that actually can be set to numeric "1" or "0"
					if (!empty($setting['defaults']['data_type']))
					{
						if ($setting['defaults']['data_type'] == 'bool')
						{
							$this->convertBoolean($newValue, is_bool($setting['defaults']['good']));
						}
						elseif($setting['defaults']['data_type'] == 'int')
						{
							$newValue = intval($newValue);

							// This is for origin_max_http_version (a toggle with possible values of 1 or 2)
							if (!empty($setting['defaults']['good']) && $setting['defaults']['good'] == 2)
							{
								$newValue++;
								if (is_string($setting['defaults']['good']))
								{
									$newValue = strval($newValue);
								}
							}
						}
					}

					if (!is_array($setting['options']['value']) && (strlen($newValue) > 0 || is_bool($newValue)) && $newValue != $setting['options']['value'])
					{
						$settingsToChange[$setting['id']] = ['id' => $setting['id'], 'value' => $newValue];
					}
				}
			}
		}
		if ($settingsToChange)
		{
			$cloudflareRepo->updateSettings($settingsToChange);
		}

		return $this->message(\XF::phrase('redirect_changes_saved_successfully'));
	}


	public function actionCache()
	{
		$this->setSectionContext('cloudflareCache');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->purgeCache();
			return $this->message(\XF::phrase('cloudflare_cache_purged'));
		}
		return $this->view('DigitalPoint\Cloudflare:Cache', 'cloudflare_purge_cache_confirm', []);
	}

	public function actionFirewall(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		$cloudflareRepo = $this->getCloudflareRepo();

		if ($this->isPost())
		{
			$rules = $cloudflareRepo->getFirewallRules();
			$enables = $this->filter('enable', 'array-bool');

			if ($rules)
			{
				$changes = false;

				foreach ($rules as $rule)
				{
					if (isset($enables[$rule['id']]) && $rule['enabled'] != $enables[$rule['id']])
					{
						$changes = true;

						if ($rule['enabled'])
						{
							$cloudflareRepo->disableFirewallRule($rule);
						}
						else
						{
							$cloudflareRepo->enableFirewallRule($rule);
						}
					}
				}

				if ($changes)
				{
					return $this->message(\XF::phrase('your_changes_have_been_saved'));
				}
			}

			return $this->message(\XF::phrase('no_changes_have_been_made'));
		}

		$viewParams = [
			'rules' => $cloudflareRepo->getFirewallRules(),
			'rules_user_agent' => $cloudflareRepo->getFirewallUserAgentRules(),
			'rules_ip' => $cloudflareRepo->getZoneFirewallAccessRules(),
			'dash_base' => $cloudflareRepo->getDashBase(),
		];

		if (!empty($viewParams['rules']))
		{
			$promises = [];

			foreach ($viewParams['rules'] as $key => $rule)
			{
				if (in_array($rule['action'], ['challenge', 'js_challenge', 'managed_challenge']))
				{
					$promises[$key]  = $cloudflareRepo->getGraphQLCaptchaSolveRate($rule['id']);
				}
				elseif (in_array($rule['action'], ['block', 'allow', 'log', 'bypass']))
				{
					$promises[$key]  = $cloudflareRepo->getGraphQLRuleActivityQuery($rule['id']);
				}
			}

			$promises = $cloudflareRepo->resolvePromises($promises);

			foreach ($viewParams['rules'] as $key => &$rule)
			{
				if (in_array($rule['action'], ['challenge', 'js_challenge', 'managed_challenge']))
				{
					$rule['captcha_solve_rate'] = $cloudflareRepo->prepareGraphQLCaptchaSolveRate($promises[$key]);
				}
				elseif (in_array($rule['action'], ['block', 'allow', 'log', 'bypass']))
				{
					$rule['captcha_solve_rate'] = $cloudflareRepo->prepareGraphQLRuleActivityQuery($promises[$key]);
				}
			}
		}

		return $this->view('DigitalPoint\Cloudflare:Firewall\List', 'cloudflare_firewall_list', $viewParams);
	}

	public function actionFirewallBlockAiScrapers(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->updateFirewallBlockAiScrapersRule();
			return $this->redirect($this->buildLink('cloudflare/firewall'));
		}

		$viewParams = [
		];

		return $this->view('DigitalPoint\Cloudflare:Firewall\BlockAiScrapers\Add', 'cloudflare_firewall_blockaiscrapers_add_confirm', $viewParams);
	}

	public function actionFirewallBlockInternal(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->updateFirewallBlockInternalRule();
			return $this->redirect($this->buildLink('cloudflare/firewall'));
		}

		$viewParams = [
		];

		return $this->view('DigitalPoint\Cloudflare:Firewall\BlockInternal\Add', 'cloudflare_firewall_blockinternal_add_confirm', $viewParams);
	}

	public function actionFirewallRegistrationChallenge(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->updateFirewallRegistrationChallengeRule($this->filter('contact', 'bool'));
			$this->repository('XF:Option')->updateOption('cloudflareRegisterOverlay', false);
			return $this->redirect($this->buildLink('cloudflare/firewall'));
		}

		$viewParams = [
		];

		return $this->view('DigitalPoint\Cloudflare:Firewall\RegistrationChallenge\Add', 'cloudflare_firewall_registrationchallenge_add_confirm', $viewParams);
	}

	public function actionFirewallCountry(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		$cloudflareRepo = $this->getCloudflareRepo();
		$blockingRule = $cloudflareRepo->getFirewallRuleByRef('xf_country_block');

		if ($this->isPost())
		{
			$countries = $this->filter('countries', 'array-str');

			if ($countries)
			{
				$cloudflareRepo->updateFirewallRuleByRef(
					'xf_country_block',
					$this->data('DigitalPoint\Cloudflare:Expression')->countriesToExpression($countries, $this->filter('apply_to', 'str')),
					\XF::phrase('xenforo_country_blocking'),
					'block'
				);
			}
			else
			{
				if ($blockingRule)
				{
					$this->getCloudflareRepo()->deleteFirewallCountryBlockingRule($blockingRule['ruleset_id'], $blockingRule['id']);
				}
			}

			return $this->redirect($this->buildLink('cloudflare/firewall'));
		}

		$viewParams = [
			'blocking_rule' => $blockingRule,
			'selected_countries' => empty($blockingRule['expression']) ? [] : $this->data('DigitalPoint\Cloudflare:Expression')->expressionToCountries($blockingRule['expression']),
			'countries' => $this->data('DigitalPoint\Cloudflare:Country')->getCountries()
		];

		return $this->view('DigitalPoint\Cloudflare:Firewall\Country', 'cloudflare_firewall_country', $viewParams);
	}

	public function actionFirewallFilterDelete(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		$identifier = $this->filter('id', 'str');

		$cloudflareRepo = $this->getCloudflareRepo();

		if ($this->isPost())
		{
			$rules = $cloudflareRepo->getFirewallRules();
			foreach ($rules as $rule)
			{
				if ($rule['id'] == $identifier)
				{
					$cloudflareRepo->deleteFirewallCountryBlockingRule($rule['ruleset_id'], $identifier);
				}
			}
			return $this->redirect($this->buildLink('cloudflare/firewall'), \XF::phrase('cloudflare_firewall_rule_deleted'));
		}

		return $this->view('DigitalPoint\Cloudflare:Firewall\Country\Delete', 'cloudflare_firewall_country_delete_confirm', ['id' => $identifier]);
	}

	public function actionFirewallIpAdd(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		if ($this->isPost())
		{
			$ip = $this->filter('ip', 'str');
			$parsed = \XF\Util\Ip::parseIpRangeString($ip);

			$country = $this->filter('country', 'str');
			$asn = $this->filter('asn', 'uint');

			if ($parsed)
			{
				if (!empty($parsed['isRange']))
				{
					$ip = \XF\Util\Ip::convertIpBinaryToString($parsed['startRange']) . '/' . $parsed['cidr'];
				}
				else
				{
					$ip = $parsed['printable'];
				}
			}
			elseif(!$country && !$asn)
			{
				return $this->error(\XF::phrase('please_enter_valid_ip_or_ip_range'));
			}

			$this->getCloudflareRepo()->addFirewallAccessRule(
				$ip,
				$asn,
				$country,
				$this->filter('mode', 'str'),
				$this->filter('note', 'str'),
				$this->filter('days', 'uint')
			);

			return $this->redirect($this->buildLink('cloudflare/firewall'));
		}

		$viewParams = [
			//		'countries' => $this->getCloudflareRepo()->getCountries()  // Not using because country blocking is an Enterprise only feature.
		];

		return $this->view('DigitalPoint\Cloudflare:Firewall\Ip\Add', 'cloudflare_firewall_ip_add', $viewParams);
	}

	public function actionFirewallIpDelete(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		$identifier = $this->filter('id', 'str');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->deleteFirewallAccessRule($identifier);
			return $this->redirect($this->buildLink('cloudflare/firewall'), \XF::phrase('cloudflare_firewall_rule_deleted'));
		}

		return $this->view('DigitalPoint\Cloudflare:Firewall\Ip\Delete', 'cloudflare_firewall_ip_delete_confirm', ['id' => $identifier]);
	}


	public function actionFirewallUseragent(ParameterBag $params)
	{
		$this->assertPostOnly();

		$cloudflareRepo = $this->getCloudflareRepo();

		$rules = $cloudflareRepo->getFirewallUserAgentRules();
		$enables = $this->filter('enable', 'array-bool');

		if ($rules)
		{
			$changes = false;

			foreach ($rules as $rule)
			{
				if (isset($enables[$rule['id']]) && $rule['paused'] == $enables[$rule['id']]) // Careful, comparing paused to enabled.
				{
					$changes = true;

					if ($rule['paused'])
					{
						$cloudflareRepo->enableFirewallUserAgentRule($rule);
					}
					else
					{
						$cloudflareRepo->disableFirewallUserAgentRule($rule);
					}
				}
			}

			if ($changes)
			{
				return $this->message(\XF::phrase('your_changes_have_been_saved'));
			}
		}

		return $this->message(\XF::phrase('no_changes_have_been_made'));
	}


	public function actionFirewallUseragentAdd(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		if ($this->isPost())
		{
			$userAgent = $this->filter('user_agent', 'str');

			$this->getCloudflareRepo()->addFirewallUserAgentRule(
				$userAgent,
				$this->filter('mode', 'str'),
				$this->filter('description', 'str')
			);

			return $this->redirect($this->buildLink('cloudflare/firewall'));
		}

		return $this->view('DigitalPoint\Cloudflare:Firewall\UserAgent\Add', 'cloudflare_firewall_useragent_add', []);
	}

	public function actionFirewallUseragentDelete(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareFirewall');

		$identifier = $this->filter('id', 'str');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->deleteFirewallUserAgentRule($identifier);
			return $this->redirect($this->buildLink('cloudflare/firewall'), \XF::phrase('cloudflare_firewall_rule_deleted'));
		}

		return $this->view('DigitalPoint\Cloudflare:Firewall\UserAgent\Delete', 'cloudflare_firewall_useragent_delete_confirm', ['id' => $identifier]);
	}


	public function actionRules(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareRules');
		$cloudflareRepo = $this->getCloudflareRepo();

		$viewParams = [
			'page_rules' => $cloudflareRepo->getPageRules(),
			'cache_rules' => $cloudflareRepo->getCacheRules(),
			'dash_base' => $cloudflareRepo->getDashBase()
		];

		return $this->view('DigitalPoint\Cloudflare:Rules\List', 'cloudflare_rules_list', $viewParams);
	}

	public function actionPageRules(ParameterBag $params)
	{
		$this->assertPostOnly();

		$cloudflareRepo = $this->getCloudflareRepo();

		$rules = $cloudflareRepo->getPageRules();
		$enables = $this->filter('enable', 'array-bool');

		if ($rules)
		{
			$changes = false;

			foreach ($rules as $rule)
			{
				if (isset($enables[$rule['id']]) && $rule['status'] != ($enables[$rule['id']] ? 'active' : 'disabled'))
				{
					$changes = true;

					if ($rule['status'] == 'disabled')
					{
						$cloudflareRepo->enablePageRule($rule);
					}
					else
					{
						$cloudflareRepo->disablePageRule($rule);
					}
				}
			}

			if ($changes)
			{
				return $this->message(\XF::phrase('your_changes_have_been_saved'));
			}
		}

		return $this->message(\XF::phrase('no_changes_have_been_made'));
	}

	public function actionPageRulesDelete(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareRules');

		$identifier = $this->filter('id', 'str');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->deletePageRule($identifier);
			return $this->redirect($this->buildLink('cloudflare/rules'));
		}

		return $this->view('DigitalPoint\Cloudflare:PageRules\Delete', 'cloudflare_pagerules_delete_confirm', ['id' => $identifier]);
	}


	public function actionCacheRulesAdd(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareRules');

		$type = $this->filter('type', 'str');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->addSpecialCacheRule($type);
			return $this->redirect($this->buildLink('cloudflare/rules'));
		}

		$viewParams = [
			'type' => $type,
		];

		return $this->view('DigitalPoint\Cloudflare:CacheRules\Add', 'cloudflare_cacherules_add_confirm', $viewParams);
	}

	public function actionCacheRules(ParameterBag $params)
	{
		$this->assertPostOnly();

		$cloudflareRepo = $this->getCloudflareRepo();

		$rules = $cloudflareRepo->getCacheRules();
		$enables = $this->filter('enable', 'array-bool');

		if ($rules)
		{
			$changes = false;

			foreach ($rules as $rule)
			{
				if (isset($enables[$rule['id']]) && $rule['enabled'] != $enables[$rule['id']])
				{
					$changes = true;

					if ($rule['enabled'] == false)
					{
						$cloudflareRepo->enableCacheRule($rule['id']);
					}
					else
					{
						$cloudflareRepo->disableCacheRule($rule['id']);
					}
				}
			}

			if ($changes)
			{
				return $this->message(\XF::phrase('your_changes_have_been_saved'));
			}
		}

		return $this->message(\XF::phrase('no_changes_have_been_made'));
	}

	public function actionCacheRulesDelete(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareRules');

		$rulesetId = $this->filter('rid', 'str');
		$identifier = $this->filter('id', 'str');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->deleteCacheRule($rulesetId, $identifier);
			return $this->redirect($this->buildLink('cloudflare/rules'));
		}

		return $this->view('DigitalPoint\Cloudflare:CacheRules\Delete', 'cloudflare_cacherules_delete_confirm', ['rid' => $rulesetId, 'id' => $identifier]);
	}


	public function actionAccess(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareAccess');

		$cloudflareRepo = $this->getCloudflareRepo();

		$viewParams = [
			'apps' => $cloudflareRepo->getAccessApps(),
			'groups' => $cloudflareRepo->getAccessGroups(),
			'dash_base' => $cloudflareRepo->getTeamsDashBase()
		];

		return $this->view('DigitalPoint\Cloudflare:Access\List', 'cloudflare_access_list', $viewParams);
	}

	public function actionAccessAdd(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareAccess');

		$type = $this->filter('type', 'str');

		if (!count($this->getCloudflareRepo()->getAccessIdentityProviders()))
		{
			return $this->error(\XF::phrase('no_cloudflare_access_login_methods', ['url' => sprintf('%s/%s', $this->getCloudflareRepo()->getZeroTrustDashBase(), 'settings/authentication')]));
		}

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->addAdminAccessApp($type);
			return $this->redirect($this->buildLink('cloudflare/access'));
		}

		$viewParams = [
			'type' => $type
		];

		return $this->view('DigitalPoint\Cloudflare:Access\Add', 'cloudflare_access_add_confirm', $viewParams);
	}

	public function actionAccessDelete(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareAccess');

		$identifier = $this->filter('id', 'str');

		if ($this->isPost())
		{
			$this->getCloudflareRepo()->deleteAccessApp($identifier);
			return $this->redirect($this->buildLink('cloudflare/access'));
		}

		return $this->view('DigitalPoint\Cloudflare:Access\Delete', 'cloudflare_access_delete_confirm', ['id' => $identifier]);
	}


	public function actionProxy(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareProxy');
		$imageWorkerId = 'xf-image-proxy';
		$urlWorkerId = 'xf-url-proxy';

		$cloudflareRepo = $this->getCloudflareRepo();

		$subdomain = $cloudflareRepo->getWorkerSubdomain();

		if (!$subdomain)
		{
			return $this->error(\XF::phrase('need_cloudflare_worker_subdomain', ['url'=> sprintf($cloudflareRepo->workerSetupUrl, $cloudflareRepo->getAccountId())]));
		}

		$viewParams = ['subdomain' => $subdomain];

		$imageWorker = $cloudflareRepo->getWorkerById($imageWorkerId);
		$urlWorker = $cloudflareRepo->getWorkerById($urlWorkerId);

		if (!$imageWorker || !$urlWorker)
		{
			if ($this->isPost())
			{
				$workerData = $this->data('DigitalPoint\Cloudflare:Worker');

				$cloudflareRepo->uploadWorker($imageWorkerId, $workerData->getImageProxyWorker());
				$cloudflareRepo->uploadWorker($urlWorkerId, $workerData->getUrlProxyWorker());

				return $this->redirect($this->buildLink('cloudflare/proxy'));
			}

			return $this->view('DigitalPoint\Cloudflare:Proxy\CreateWorker', 'cloudflare_proxy_create_worker', $viewParams);
		}

		$cloudflareRepo->workerSubdomainEnable($imageWorkerId);
		$cloudflareRepo->workerSubdomainEnable($urlWorkerId);

		if ($this->isPost())
		{
			$options = [
				'cfWorkersSubdomain' => '',
				'cfProxy' => [
					'image' => false,
					'url' => false,
				]
			];

			if ($this->filter('image', 'bool'))
			{
				$options['cfProxy']['image'] = true;
			}
			if ($this->filter('url', 'bool'))
			{
				$options['cfProxy']['url'] = true;
			}

			if ($options['cfProxy']['image'] || $options['cfProxy']['url'])
			{
				$options['cfWorkersSubdomain'] = $subdomain;
			}

			$this->getOptionRepo()->updateOptions($options);

			return $this->message(\XF::phrase('redirect_changes_saved_successfully'));
		}

		return $this->view('DigitalPoint\Cloudflare:ImageProxy', 'cloudflare_image_proxy', $viewParams);
	}

	public function actionEdgeCaching(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareEdgeCaching');

		$viewParams = [
			'action' => $this->filter('action', 'str'),
		];

		$type = $this->filter('type', 'str');

		if ($type == 'media')
		{
			if ($viewParams['action'] == 'enable')
			{
				if ($this->isPost())
				{
					$this->getOptionRepo()->updateOption('cfMediaCachingSeconds', $this->filter('seconds', 'uint'));
					$this->getCloudflareRepo()->addSpecialCacheRule('media_cache');

					return $this->redirect($this->buildLink('cloudflare/edge-caching'));
				}

				return $this->view('DigitalPoint\Cloudflare:MediaCaching\Enable', 'cloudflare_media_caching_enable', $viewParams);

			}
			elseif ($viewParams['action'] == 'disable')
			{
				if ($this->isPost())
				{
					$this->getOptionRepo()->updateOption('cfMediaCachingSeconds', 0);
					$this->getCloudflareRepo()->deleteSpecialCacheRule('media_cache');

					return $this->redirect($this->buildLink('cloudflare/edge-caching'));
				}

				return $this->view('DigitalPoint\Cloudflare:MediaCaching\Disable', 'cloudflare_media_caching_disable_confirm', $viewParams);
			}
		}
		elseif ($type == 'guest_page')
		{
			if ($viewParams['action'] == 'enable')
			{
				if ($this->isPost())
				{
					$this->getOptionRepo()->updateOption('cfPageCachingSeconds', $this->filter('seconds', 'uint'));
					$this->getCloudflareRepo()->addSpecialCacheRule('guest_cache');

					return $this->redirect($this->buildLink('cloudflare/edge-caching'));
				}

				return $this->view('DigitalPoint\Cloudflare:GuestCaching\Enable', 'cloudflare_guest_caching_enable', $viewParams);

			}
			elseif ($viewParams['action'] == 'disable')
			{
				if ($this->isPost())
				{
					$this->getOptionRepo()->updateOption('cfPageCachingSeconds', 0);
					$this->getCloudflareRepo()->deleteSpecialCacheRule('guest_cache');

					return $this->redirect($this->buildLink('cloudflare/edge-caching'));
				}

				return $this->view('DigitalPoint\Cloudflare:GuestCaching\Disable', 'cloudflare_guest_caching_disable_confirm', $viewParams);
			}
		}

		return $this->view('DigitalPoint\Cloudflare:GuestCaching\Index', 'cloudflare_edge_caching', $viewParams);
	}

	public function actionGuestCachingEdit(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareEdgeCaching');

		if (!$this->options()->cfPageCachingSeconds)
		{
			throw $this->exception($this->error(\XF::phrase('guest_page_caching_not_enabled')));
		}
		if ($this->isPost())
		{
			$this->getOptionRepo()->updateOption('cfPageCachingSeconds', $this->filter('seconds', 'uint'));

			return $this->redirect($this->buildLink('cloudflare/edge-caching'));
		}

		return $this->view('DigitalPoint\Cloudflare:GuestCaching\Edit', 'cloudflare_guest_caching_edit', []);
	}

	public function actionMediaCachingEdit(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareEdgeCaching');

		if (!$this->options()->cfMediaCachingSeconds)
		{
			throw $this->exception($this->error(\XF::phrase('media_attachment_caching_not_enabled')));
		}
		if ($this->isPost())
		{
			$this->getOptionRepo()->updateOption('cfMediaCachingSeconds', $this->filter('seconds', 'uint'));

			return $this->redirect($this->buildLink('cloudflare/edge-caching'));
		}

		return $this->view('DigitalPoint\Cloudflare:MediaCaching\Edit', 'cloudflare_media_caching_edit', []);
	}

	public function actionDmarc(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareDmarc');

		$cloudflareRepo = $this->getCloudflareRepo();
		$dmarcReport = $cloudflareRepo->getDmarcReports();
		$cloudflareRepo->getZoneId($hostname, true, true);

		if (!empty($dmarcReport['result']) && !empty($dmarcReport['result']['status']) && $dmarcReport['result']['status'] == 'missing-dmarc-report')
		{
			return $this->error(\XF::phrase('no_rua_in_dmarc_record', ['url' => sprintf($cloudflareRepo->getDashBaseAccount() . '/%s/email/dmarc-management/dmarc-reports/dmarc/wizard', $hostname)]));
		}

		$approvedSources = [];

		if (!empty($dmarcReport['result']['approved_sources']))
		{
			foreach($dmarcReport['result']['approved_sources'] as $source)
			{
				$approvedSources[] = $source['slug'];
			}
		}

		$dmarcSources = $cloudflareRepo->resolvePromises([$cloudflareRepo->getDmarcSources($approvedSources)]);

		$viewParams = [
			'sources' => $cloudflareRepo->prepareGraphQLDmarcSources($dmarcSources[0]),
			'management_url' => sprintf($cloudflareRepo->getDashBaseAccount() . '/%s/email/dmarc-management', $hostname)
		];

		return $this->view('DigitalPoint\Cloudflare:Dmarc\Source', 'cloudflare_dmarc_management', $viewParams);
	}

	public function actionR2(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareR2');
		$this->assertHasToken();
		$cloudflareRepo = $this->getCloudflareRepo();

		// skip check if we've already specified a bucket. This is a class A operation, so don't do it unless it's necessary (if no buckets have been defined, or we are enabling a bucket and need bucket list for drop-down).
		$checkR2Capability = true;
		foreach($this->options()->cfR2Bucket as $bucket)
		{
			if ($bucket)
			{
				$checkR2Capability = false;
				break;
			}
		}

		if ($checkR2Capability)
		{
			$buckets = $cloudflareRepo->getR2Buckets();

			if ($buckets === false)
			{
				return $this->error(\XF::phrase('need_cloudflare_r2', ['url'=> $cloudflareRepo->getDashBaseAccount() . '/r2/overview/plans']));
			}
		}

		$viewParams = [
			'action' => $this->filter('action', 'str'),
			'type' => $this->filter('type', 'str'),
		];

		if ($viewParams['action'] == 'enable')
		{
			if (!$checkR2Capability)
			{
				$buckets = $cloudflareRepo->getR2Buckets();
			}

			$viewParams['buckets'] = $buckets;

			if ($this->isPost())
			{
				$bucket = $this->filter('bucket', 'str');

				if (!preg_match('/^[a-z0-9][a-z0-9-]{1,61}[a-z0-9]$/', $bucket))
				{
					return $this->error(\XF::phrase('bucket_name_must_be'));
				}

				// Normally shouldn't be needed, but just in case the token changed in some way outside of the normal options, this will do a double-check.
				$cloudflareRepo->verifyToken();

				if(empty($buckets[$bucket]))
				{
					$cloudflareRepo->addR2Bucket($bucket);
				}

				$publicSubdomain = $this->filter('public_subdomain', 'str');

				if ($publicSubdomain)
				{
					$cloudflareRepo->addR2CustomDomain($bucket, $publicSubdomain);
					$this->getOptionRepo()->updateOption('cfExternalDataUrl', 'https://' . $publicSubdomain);

					if ($this->filter('create_cache_rule', 'bool'))
					{
						$this->getCloudflareRepo()->addSpecialCacheRule('r2-bucket', $publicSubdomain);
					}
				}

				$r2Bucket = \XF::options()->cfR2Bucket;
				$r2Bucket[$viewParams['type']] = $bucket;

				$this->getOptionRepo()->updateOption('cfR2Bucket', $r2Bucket);

				return $this->redirect($this->buildLink('cloudflare/r2'));
			}

			if ($viewParams['type'] == 'data')
			{
				$viewParams['public_subdomain'] = 'data.' . parse_url($this->options()->boardUrl, PHP_URL_HOST);
			}

			return $this->view('DigitalPoint\Cloudflare:R2\Enable', 'cloudflare_r2_enable', $viewParams);

		}
		elseif ($viewParams['action'] == 'disable')
		{
			if ($this->isPost())
			{
				$r2Bucket = \XF::options()->cfR2Bucket;
				$r2Bucket[$viewParams['type']] = '';

				$this->getOptionRepo()->updateOption('cfR2Bucket', $r2Bucket);

				return $this->redirect($this->buildLink('cloudflare/r2'));
			}

			return $this->view('DigitalPoint\Cloudflare:R2\Disable', 'cloudflare_r2_disable_confirm', $viewParams);
		}

		$accountStats = $cloudflareRepo->getR2AccountInfo();
		$allPromises = $accountStats;

		$bucketStats = [];
		foreach($this->options()->cfR2Bucket as $bucketName)
		{
			if ($bucketName)
			{
				$bucketStats = array_merge($bucketStats, $cloudflareRepo->getR2BucketInfo($bucketName));
			}
		}

		$allPromises = array_merge($allPromises, $bucketStats);
		$allPromises = $cloudflareRepo->resolvePromises($allPromises);

		foreach ($accountStats as $key => $item)
		{
			$accountStats[$key] = $allPromises[$key];
		}

		$bucketStats = [];
		foreach($this->options()->cfR2Bucket as $bucketName)
		{
			if ($bucketName)
			{
				foreach ($allPromises as $key => $item)
				{
					if (substr($key, 0, strlen($bucketName) + 1) === $bucketName . ':')
					{
						$itemKey = substr($key, strlen($bucketName) + 1);

						$bucketStats[$bucketName][$itemKey] = $item;
					}
				}
			}
		}

		foreach ($bucketStats as $key => $bucket)
		{
			$bucketStats[$key] = $cloudflareRepo->parseR2BucketInfo($bucket);
		}

		$viewParams['account_stats'] = $cloudflareRepo->parseR2AccountInfo($accountStats);
		$viewParams['bucket_stats'] = $bucketStats;
		$viewParams['dash_base_account'] = $cloudflareRepo->getDashBaseAccount();

		// Debug to fake that we have a Cloudflare subscription on zone even if we don't.
		//$viewParams['account_stats']['has_subscription'] = 1;

		return $this->view('DigitalPoint\Cloudflare:R2', 'cloudflare_r2', $viewParams);
	}

	public function actionR2Log(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareR2');
		$this->assertHasToken();
		$cloudflareRepo = $this->getCloudflareRepo();

		$action = $this->filter('action', 'str');
		$type = $this->filter('type', 'str');

		$bucket = (!empty($this->options()->cfR2Bucket[$type]) ? $this->options()->cfR2Bucket[$type] : null);

		$viewParams = [
			'logs' => $cloudflareRepo->getGraphQLR2OperationsLog($bucket, $action)
		];

		return $this->view('DigitalPoint\Cloudflare:R2\Log', 'cloudflare_r2_log', $viewParams);
	}

	public function actionR2TokenAuth(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareR2');
		$this->assertHasToken();
		$cloudflareRepo = $this->getCloudflareRepo();

		// Will throw error if there is no subscription.
		$cloudflareRepo->getZoneSubscription();

		if (empty($this->options()->cfR2Bucket['internal-data/attachments']))
		{
			return $this->error(\XF::phrase('no_bucket_for_x'));
		}
		$bucket = $this->options()->cfR2Bucket['internal-data/attachments'];

		$action = $this->filter('action', 'str');

		if ($action === 'disable')
		{
			if ($this->isPost())
			{
				$publicSubdomain = parse_url($this->options()->cfExternalAttachmentUrl, PHP_URL_HOST);

				$cloudflareRepo->deleteFirewallRuleByRef('r2_token_auth');
				$cloudflareRepo->removeR2CustomDomain($bucket, $publicSubdomain);

				$this->getOptionRepo()->updateOptions([
					'cfExternalAttachmentUrl' => '',
					'cfSecretKey' => ''
				]);

				return $this->redirect($this->buildLink('cloudflare/r2'));
			}

			$viewParams = [];

			return $this->view('DigitalPoint\Cloudflare:R2\TokenAuth\Disable', 'cloudflare_r2_token_auth_disable_confirm', $viewParams);
		}
		else
		{
			if ($this->isPost())
			{
				//		$bucket = $this->filter('bucket', 'str');

				//		if (!preg_match('/^[a-z0-9][a-z0-9-]{1,61}[a-z0-9]$/', $bucket))
				//		{
				//			return $this->error(\XF::phrase('bucket_name_must_be'));
				//		}

				// Normally shouldn't be needed, but just in case the token changed in some way outside of the normal options, this will do a double-check.
				$cloudflareRepo->verifyToken();

				//		if(empty($buckets[$bucket]))
				//		{
				//			$cloudflareRepo->addR2Bucket($bucket);
				//		}

				$publicSubdomain = $this->filter('public_subdomain', 'str');

				if (empty($publicSubdomain))
				{
					return $this->error(\XF::phrase('missing_subdomain'));
				}

				$secretKey = \XF::generateRandomString(32);

				// Doing this first in case it fails (can't do it on a Free subscription)
				$cloudflareRepo->updateFirewallR2TokenAuth($publicSubdomain, $secretKey);

				$cloudflareRepo->addR2CustomDomain($bucket, $publicSubdomain);

				$this->getOptionRepo()->updateOptions([
					'cfExternalAttachmentUrl' => 'https://' . $publicSubdomain,
					'cfSecretKey' => $secretKey
				]);

				return $this->redirect($this->buildLink('cloudflare/r2'));
			}

			$viewParams = [];

			$viewParams['public_subdomain'] = 'attachments.' . parse_url($this->options()->boardUrl, PHP_URL_HOST);

			return $this->view('DigitalPoint\Cloudflare:R2\TokenAuth\Enable', 'cloudflare_r2_token_auth_enable', $viewParams);
		}


	}

	public function actionToolsRequestTrace()
	{
		$this->setSectionContext('cloudflareToolRequestTrace');
		$this->assertHasToken();

		if ($this->isPost())
		{
			$url = $this->filter('url', 'str');
			$method = $this->filter('method', 'str');
			$protocol = $this->filter('protocol', 'str');
			$botScore = $this->filter('bot_score', 'uint');
			$country = $this->filter('country', 'str');
			$skipChallenge = $this->filter('skip_challenge', 'bool');

			// check if it exists first because 0 and empty are different here
			if (isset($_POST['threat_score']))
			{
				$threatScore = $this->filter('threat_score', 'uint');
			}
			else
			{
				$threatScore = null;
			}

			$viewParams = [
				'result' => $this->getCloudflareRepo()->requestTrace($url, $method, $protocol, $botScore, $country, $skipChallenge, $threatScore)
			];

			return $this->view('DigitalPoint\Cloudflare:Tools\RequestTrace\Result', 'cloudflare_tools_request_trace_result', $viewParams);
		}

		$viewParams = [
			'countries' => $this->data('DigitalPoint\Cloudflare:Country')->getCountries()
		];
		return $this->view('DigitalPoint\Cloudflare:Tools\RequestTrace', 'cloudflare_tools_request_trace', $viewParams);
	}

	public function actionToolsIpDetails()
	{
		$this->setSectionContext('cloudflareToolIpDetails');
		$this->assertHasToken();

		if ($this->isPost())
		{
			$ip = $this->filter('ip', 'str');

			$viewParams = [
				'result' => $this->getCloudflareRepo()->ipDetails($ip),
			];

			$countries = $this->data('DigitalPoint\Cloudflare:Country')->getCountries();
			if (!empty($viewParams['result']['result'][0]['belongs_to_ref']['country']) && !empty($countries[$viewParams['result']['result'][0]['belongs_to_ref']['country']]))
			{
				$viewParams['country'] = $countries[$viewParams['result']['result'][0]['belongs_to_ref']['country']];
			}

			return $this->view('DigitalPoint\Cloudflare:Tools\IpDetails\Result', 'cloudflare_tools_ip_details_result', $viewParams);
		}

		$viewParams = [
			'ip' => \XF::app()->request()->getIp()
		];
		return $this->view('DigitalPoint\Cloudflare:Tools\IpDetails', 'cloudflare_tools_ip_details', $viewParams);
	}

	public function actionToolsDomainDetails()
	{
		$this->setSectionContext('cloudflareToolDomainDetails');
		$this->assertHasToken();

		if ($this->isPost())
		{
			$domain = $this->filter('domain', 'str');

			$viewParams = [
				'result' => $this->getCloudflareRepo()->domainDetails($domain)
			];

			return $this->view('DigitalPoint\Cloudflare:Tools\DomainDetails\Result', 'cloudflare_tools_domain_details_result', $viewParams);
		}

		$viewParams = [
			'hostname' => parse_url($this->options()->boardUrl, PHP_URL_HOST)
		];
		return $this->view('DigitalPoint\Cloudflare:Tools\DomainDetails', 'cloudflare_tools_domain_details', $viewParams);
	}

	public function actionToolsWhois()
	{
		$this->setSectionContext('cloudflareToolWhois');
		$this->assertHasToken();

		if ($this->isPost())
		{
			$domain = $this->filter('domain', 'str');

			$viewParams = [
				'result' => $this->getCloudflareRepo()->whois($domain)
			];

			return $this->view('DigitalPoint\Cloudflare:Tools\Whois\Result', 'cloudflare_tools_whois_result', $viewParams);
		}

		$viewParams = [
			'hostname' => parse_url($this->options()->boardUrl, PHP_URL_HOST)
		];
		return $this->view('DigitalPoint\Cloudflare:Tools\Whois', 'cloudflare_tools_whois', $viewParams);
	}

	// This is Enterprise-only, so not exposing it for now.
	public function actionToolsTraceroute()
	{
		$this->setSectionContext('cloudflareTools');
		$this->assertHasToken();

		if ($this->isPost())
		{
			$hostname = $this->filter('hostname', 'str');
			$maxTTL = $this->filter('max_ttl', 'uint');
			$packetType = $this->filter('packet_type', 'str');
			$packetsPerTTL = $this->filter('packets_per_ttl', 'uint');
			$port = $this->filter('port', 'uint');
			$waitTime = $this->filter('wait_time', 'posint');

			$viewParams = [
				'result' => $this->getCloudflareRepo()->traceroute([$hostname], [], [
					'max_ttl' => $maxTTL,
					'packet_type' => $packetType,
					'packets_per_ttl' => $packetsPerTTL,
					'port' => $port,
					'wait_time' => $waitTime
				])
			];
			
			return $this->view('DigitalPoint\Cloudflare:Tools\Traceroute\Result', 'cloudflare_tools_traceroute_result', $viewParams);
		}

		$viewParams = [
			'hostname' => parse_url($this->options()->boardUrl, PHP_URL_HOST)
		];
		return $this->view('DigitalPoint\Cloudflare:Tools\Traceroute', 'cloudflare_tools_traceroute', $viewParams);
	}

	public function actionTurnstileWidgetAdd()
	{
		$cloudflareRepo = $this->getCloudflareRepo();

		$site = $cloudflareRepo->getTurnstileWidgetByDomain();
		if ($site)
		{
			return $this->error(\XF::phrase('cloudflare_turnstile_site_exists', [
				'url' => $cloudflareRepo->getTurnstileSiteUrlEdit($site['sitekey'])
			]));
		}

		if ($this->isPost())
		{
			$options = $this->options();
			$result = $cloudflareRepo->addTurnstileSite(
				$options->boardTitle,
				parse_url($options->boardUrl, PHP_URL_HOST)
			);

			if (!empty($result['result']))
			{
				$this->getOptionRepo()->updateOptions([
					'captcha' => 'Turnstile',
					'extraCaptchaKeys' => [
						'turnstileSiteKey' => $result['result']['sitekey'],
						'turnstileSecretKey' => $result['result']['secret']
					]
				]);
			}

			return $this->redirect($this->app()->request()->getReferrer() . '#captcha', \XF::phrase('redirect_changes_saved_successfully'));
		}

		return $this->view('DigitalPoint\Cloudflare:Turnstile\Site\Add\Confirm', 'cloudflare_turnstile_site_add_confirm', []);
	}

	public function actionBackup()
	{
		$this->setSectionContext('cloudflareBackup');
		$this->assertCanZip();

		if ($this->isPost())
		{
			$upload = $this->request->getFile('upload', false, false);

			$zipArchive = new \ZipArchive();
			$zipArchive->open($upload->getTempFile());
			$backup = json_decode($zipArchive->getFromName('backup.json'), true);
			$zipArchive->close();

			if (!isset($backup['firewall_rules']) || !isset($backup['access_apps']) || !isset($backup['page_rules']) || !isset($backup['cache_rules']) || !isset($backup['firewall_access_rules']) || !isset($backup['firewall_useragent_rules']))
			{
				return $this->error(\XF::phrase('invalid_backup_archive'));
			}

			$this->getCloudflareRepo()->restoreBackup($backup);

			return $this->message(\XF::phrase('backup_restored'));
		}

		return $this->view('DigitalPoint\Cloudflare:Backup\View', 'cloudflare_backup');
	}

	public function actionBackupDownload(ParameterBag $params)
	{
		$this->setSectionContext('cloudflareBackup');
		$this->assertCanZip();

		$this->setResponseType('raw');

		$viewParams = [
			'file' => $this->getCloudflareRepo()->getBackup(),
		];
		return $this->view('DigitalPoint\Cloudflare:Cloudflare\Backup\Download', '', $viewParams);
	}

	protected function convertBoolean(&$bool, $trueBoolean = false)
	{
		if ($bool === '1')
		{
			$bool = ($trueBoolean ? true : 'on');
		}
		elseif(!$bool)
		{
			$bool = ($trueBoolean ? false : 'off');
		}
	}

	/**
	 * @return \DigitalPoint\Cloudflare\Repository\Cloudflare
	 */
	protected function getCloudflareRepo()
	{
		return $this->repository('DigitalPoint\Cloudflare:Cloudflare');
	}

	/**
	 * @return \XF\Repository\Option
	 */
	protected function getOptionRepo(): \XF\Repository\Option
	{
		return $this->repository('XF:Option');
	}
}