<?php

namespace DigitalPoint\Cloudflare\League\Flysystem\Adapter;

use League\Flysystem\Adapter\AbstractAdapter;
use League\Flysystem\AdapterInterface;
use League\Flysystem\Config;
use League\Flysystem\FileNotFoundException;
use function strlen;

class R2 extends AbstractAdapter
{
	protected $apiClass = null;
	/**
	 * @return \DigitalPoint\Cloudflare\Api\Cloudflare
	 */
	protected function getApiClass()
	{
		if (!$this->apiClass)
		{
			$className = 'DigitalPoint\Cloudflare\Api\Cloudflare';
			if (method_exists('XF', 'extendClass'))
			{
				$className = \XF::extendClass($className);
			}
			$this->apiClass = new $className();
		}
		return $this->apiClass;
	}

	protected $overrideBucket = null;

	public static function getAdapter($path = '', $overrideBucket = null)
	{
		return new self($path, $overrideBucket);
	}

	/**
	 * Passes through to Flysystem's disable_asserts option for just this adapter.
	 * Asserting that a file exists before you do things like read it is an extra API call for most everything.
	 */
	public $disableAsserts = true;

	public function __construct($path, $overrideBucket = null)
	{
		$this->overrideBucket = $overrideBucket;
		$this->setPathPrefix($path);
	}

	public function setPathPrefix($prefix)
	{
		$prefix = (string)$prefix;

		if ($prefix === '') {
			$this->pathPrefix = null;

			return;
		}

		$this->pathPrefix = trim($prefix);
	}

	public function getBucketByPathPrefix()
	{
		$buckets = \XF::options()->cfR2Bucket;

		if(!empty($buckets[$this->getPathPrefix()]))
		{
			return $buckets[$this->getPathPrefix()];
		}

		if ($this->overrideBucket)
		{
			return $this->overrideBucket;
		}

		return false;
	}

	public function has($path)
	{
		try {
			$this->getMetadata($path);
		}
		catch(FileNotFoundException $e)
		{
			return false;
		}
		return true;
	}

	public function write($path, $contents, Config $config)
	{
		$this->getApiClass()->createR2Object($this->getBucketByPathPrefix(), $path, $contents);
		$size = strlen($contents);
		$type = 'file';
		$exists = true;
		return compact('exists', 'contents', 'type', 'size', 'path');
	}

	public function writeStream($path, $resource, Config $config)
	{
		return $this->write($path, stream_get_contents($resource), $config);
	}

	public function readStream($path)
	{
		$data = $this->read($path);
		$stream = fopen('php://temp', 'w+b');
		fwrite($stream, $data['contents']);
		rewind($stream);

		return ['exists' => true, 'type' => 'file', 'path' => $path, 'stream' => $stream, 'timestamp' => $data['timestamp']];
	}

	public function updateStream($path, $resource, Config $config)
	{
		return $this->writeStream($path, $resource, $config);
	}

	public function update($path, $contents, Config $config)
	{
		return $this->write($path, $contents, $config);
	}

	public function read($path)
	{
		$object = $this->getApiClass()->getR2Object($this->getBucketByPathPrefix(), $path);

		if (empty($object['exists']))
		{
			throw new FileNotFoundException($path);
		}

		return ['exists' => true, 'type' => 'file', 'path' => $path, 'contents' => $object['content'], 'timestamp' => $object['timestamp']];
	}

	public function rename($path, $newpath)
	{
		$success = $this->copy($path, $newpath);
		if ($success)
		{
			$this->delete($path);
		}
		return $success;
	}

	public function copy($path, $newpath)
	{
		$bucketName = $this->getBucketByPathPrefix();
		$results = $this->getApiClass()->copyR2Object($bucketName, $newpath, $bucketName . '/' . $path);
		return $results['exists'];
	}

	public function delete($path)
	{
		$this->getApiClass()->deleteR2Object($this->getBucketByPathPrefix(), $path);
		return true;
	}

	public function listContents($directory = '', $recursive = false)
	{
		$api = $this->getApiClass();

		$objects = [];
		$continuationToken = '';
		while (true)
		{
			$result = $api->listR2Bucket($this->getBucketByPathPrefix(), $continuationToken, $directory, ($directory && !$recursive ? '/' : ''));

			$xml = new \SimpleXMLElement($result['content']);

			foreach ($xml->Contents as $file)
			{
				$objects[] = [
					'type' => 'file',
					'path' => urldecode($file->Key),
					'timestamp' => strtotime((string)$file->LastModified),
					'size' => (string)$file->Size
				];
			}

			if (empty($xml->IsTruncated) || empty($xml->NextContinuationToken))
			{
				break;
			}
			$continuationToken = (string)$xml->NextContinuationToken;
		}

		return $objects;
	}

	public function getMetadata($path)
	{
		$object = $this->getApiClass()->headR2Object($this->getBucketByPathPrefix(), $path);
		if (empty($object['exists']))
		{
			// Not a "file, check if it's a "directory"
			if (strpos($path, '.') === false)
			{
				$object = $this->getApiClass()->listR2Bucket($this->getBucketByPathPrefix(), '', $path, '/');
				if (!empty($object['content']))
				{
					$xml = new \SimpleXMLElement($object['content']);
					if (!empty($xml->KeyCount))
					{
						return ['exists' => true, 'path' => $path, 'type' => 'dir'];
					}
				}
			}
			throw new FileNotFoundException($path);
		}

		return array_merge(['path' => $path, 'type' => 'file'], $object);
	}

	public function getSize($path)
	{
		$results = $this->getMetadata($path);
		return ['path' => $path, 'type' => 'file', 'size' => empty($results['size']) ? null : $results['size']];
	}

	public function getMimetype($path)
	{
		$results = $this->getMetadata($path);
		return ['path' => $path, 'type' => 'file', 'mimetype' => empty($results['mimetype']) ? null : $results['mimetype']];
	}

	public function getTimestamp($path)
	{
		$results = $this->getMetadata($path);
		return ['path' => $path, 'type' => 'file', 'timestamp' => empty($results['timestamp']) ? null : $results['timestamp']];
	}

	public function getVisibility($path)
	{
		$visibility = AdapterInterface::VISIBILITY_PUBLIC;
		return compact('path', 'visibility');
	}

	public function setVisibility($path, $visibility)
	{
		return $this->getVisibility($path);
		return compact('path', 'visibility');
	}

	public function createDir($dirname, Config $config)
	{
		return ['path' => $dirname, 'type' => 'dir'];
	}

	public function deleteDir($dirname)
	{
		return true;
	}
}