<?php
namespace DigitalPoint\Cloudflare\Repository\Advanced;

abstract class Cloudflare extends \DigitalPoint\Cloudflare\Repository\CloudflareAbstract
{
	protected $cacheRules = null;
	protected $pageRules = null;

	protected $workers = null;

	abstract protected function getAdminEmails();
	abstract protected function getAdminAreas($include = '');
	abstract protected function addCron($jobKey, $expireDays, $id);
	abstract protected function cancelCron($jobKey);
	public $workerSetupUrl = 'https://dash.cloudflare.com/%s/workers/subdomain';

	public function getGraphQLR2OperationsLog($bucketName = null, $actionType = 'classA', $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		$results = $api->getGraphQLR2OperationsLog($accountId, $bucketName, $actionType);

		$results = $this->resolvePromises([$results])[0];

		$return = [];

		if (!empty($results['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups']))
		{
			foreach ($results['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'] as $entry)
			{
				$entry['dimensions']['time'] = strtotime($entry['dimensions']['datetime']);
				unset($entry['dimensions']['__typename'], $entry['dimensions']['datetime']);
				$return[] = $entry['dimensions'];
			}
		}

		return $return;
	}

	public function getFirewallRuleByRef(string $reference, $hostname = null)
	{
		$rules = $this->getFirewallRules($hostname);
		if ($rules)
		{
			foreach ($rules as $rule)
			{
				if (!empty($rule['ref']) && $rule['ref'] == $reference)
				{
					return $rule;
				}
			}
		}
		return [];
	}

	public function updateFirewallBlockAiScrapersRule($hostname = null)
	{
		$dataClass = $this->getClassName('DigitalPoint\Cloudflare\Data\Expression');
		$dataClass = new $dataClass();
		$expression = $dataClass->getBlockAiScrapers();

		$this->updateFirewallRuleByRef(
			'',
			$expression,
			$this->phrase('ai_scrapers_and_crawlers_rule'),
			'block'
		);
	}

	public function updateFirewallBlockInternalRule($hostname = null)
	{
		$dataClass = $this->getClassName('DigitalPoint\Cloudflare\Data\Expression');
		$dataClass = new $dataClass();
		$expression = $dataClass->getBlockInternal();

		$this->updateFirewallRuleByRef(
			'xf_internal_block',
			$expression,
			$this->phrase('xenforo_internal_directory_blocking'),
			'block'
		);
	}

	public function updateFirewallR2TokenAuth($subdomain, $secretKey)
	{
		$dataClass = $this->getClassName('DigitalPoint\Cloudflare\Data\Expression');
		$dataClass = new $dataClass();
		$expression = $dataClass->getR2TokenAuth($subdomain, $secretKey);

		$this->updateFirewallRuleByRef(
			'r2_token_auth',
			$expression,
			$this->phrase('r2_token_authentication'),
			'block'
		);
	}

	public function deleteFirewallRuleByRef($reference)
	{
		$rule = $this->getFirewallRuleByRef($reference);
		if ($rule)
		{
			$this->deleteFirewallRule($rule['ruleset_id'], $rule['id']);
		}
	}

	public function deleteFirewallRule($rulesetId, $identifier)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();
		$api->deleteFirewallRule($zoneId, $rulesetId, $identifier);
	}

	public function enableFirewallRule(array $rule)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updateFirewallRule(
			$zoneId,
			$rule['ruleset_id'],
			$rule['id'],
			$rule['action'],
			$rule['expression'],
			$rule['description'],
			true,
			$rule['last_updated'],
			$rule['version'],
			@$rule['ref']
		);
	}

	public function disableFirewallRule(array $rule)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updateFirewallRule(
			$zoneId,
			$rule['ruleset_id'],
			$rule['id'],
			$rule['action'],
			$rule['expression'],
			$rule['description'],
			false,
			$rule['last_updated'],
			$rule['version'],
			@$rule['ref']
		);
	}

	public function updateFirewallRegistrationChallengeRule($includeContact, $hostname = null)
	{
		$dataClass = $this->getClassName('DigitalPoint\Cloudflare\Data\Expression');
		$dataClass = new $dataClass();
		$expression = $dataClass->getRegistrationChallenge($includeContact);

		$this->updateFirewallRuleByRef(
			'xf_registration_challenge',
			$expression,
			$this->phrase('xenforo_registration_contact_challenge'),
			'managed_challenge'
		);
	}


	public function updateFirewallRuleByRef($reference, $expression, $description, $action = 'block', $hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$ruleset = $api->getFirewallRules($zoneId);

		if (!empty($ruleset['result']['rules']))
		{
			foreach($ruleset['result']['rules'] as $rule)
			{
				if ($reference == $rule['ref'])
				{
					$api->updateFirewallRule(
						$zoneId,
						$ruleset['result']['id'],
						$rule['id'],
						$action,
						$expression,
						$description,
						$rule['enabled'],
						$rule['last_updated'],
						$rule['version'],
						$reference
					);
					return;
				}
			}
		}

		$phase = $api->getRulesetPhase($zoneId, 'http_request_firewall_custom/entrypoint', 404);
		if ($phase && !empty($phase['result']['id']))
		{
			$api->createFirewallRule($zoneId, $phase['result']['id'], $expression, $description, $reference, true, $action);
		}
		else
		{
			$api->createFirewallRule($zoneId, '', $expression, $description, $reference, true, $action);
		}
	}

	public function deleteFirewallCountryBlockingRule($rulesetId, $identifier, $hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();
		$api->deleteFirewallRule($zoneId, $rulesetId, $identifier);
	}

	public function addFirewallAccessRule($ip, $asn, $country, $mode, $notes = '', $expireDays = 0, $hostname = '')
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		if (substr_count($ip, '/'))
		{
			$configuration = [
				'target' => 'ip_range',
				'value' => $ip
			];
		}
		elseif(substr_count($ip, ':'))
		{
			$configuration = [
				'target' => 'ip6',
				'value' => $ip
			];
		}
		elseif(substr_count($ip, '.'))
		{
			$configuration = [
				'target' => 'ip',
				'value' => $ip
			];
		}
		elseif($asn)
		{
			$configuration = [
				'target' => 'asn',
				'value' => 'AS' . $asn
			];
		}
		elseif($country)
		{
			$configuration = [
				'target' => 'country',
				'value' => $country
			];
		}

		$result = $api->createFirewallAccessRule($zoneId, $configuration, $mode, $notes);

		if (!empty($result['result']['id']))
		{
			if ($expireDays)
			{
				$id = $result['result']['id'];
				$this->addCron('cfDelRule_' . $id, $expireDays, $id);
			}

			return $result['result']['id'];
		}
		return false;
	}

	public function deleteFirewallAccessRule($identifier, $hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();
		$api->deleteFirewallAccessRule($zoneId, $identifier);

		// Cancel any future queued job to expire it down the road
		$this->cancelCron('cfDelRule_' . $identifier);
	}



	public function addFirewallUserAgentRule($userAgent, $mode, $description = '', $hostname = '')
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$configuration = [
			'target' => 'ua',
			'value' => $userAgent
		];

		$api->createFirewallUserAgentRule($zoneId, $configuration, $mode, $description);
	}

	public function enableFirewallUserAgentRule(array $rule)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updateFirewallUserAgentRule(
			$zoneId,
			$rule['id'],
			$rule['mode'],
			$rule['configuration'],
			$rule['description'],
			false
		);
	}

	public function disableFirewallUserAgentRule(array $rule)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updateFirewallUserAgentRule(
			$zoneId,
			$rule['id'],
			$rule['mode'],
			$rule['configuration'],
			$rule['description'],
			true
		);
	}

	public function deleteFirewallUserAgentRule($identifier, $hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();
		$api->deleteFirewallUserAgentRule($zoneId, $identifier);
	}



	public function getPageRule($identifier)
	{
		if (!$this->pageRules)
		{
			$this->pageRules = $this->getPageRules();
		}

		$rule = null;

		if ($this->pageRules)
		{
			foreach ($this->pageRules as $rule)
			{
				if ($rule['id'] == $identifier)
				{
					break;
				}
			}
		}

		return $rule;
	}


	public function enablePageRule($rule)
	{
		if (!is_array($rule))
		{
			$rule = $this->getPageRule($rule);
		}

		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updatePageRule(
			$zoneId,
			$rule['id'],
			'active',
			$rule['targets'],
			$rule['actions'],
			$rule['priority']
		);
	}

	public function disablePageRule($rule)
	{
		if (!is_array($rule))
		{
			$rule = $this->getPageRule($rule);
		}

		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updatePageRule(
			$zoneId,
			$rule['id'],
			'disabled',
			$rule['targets'],
			$rule['actions'],
			$rule['priority']
		);
	}

	public function deletePageRule($identifier, $hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();
		$api->deletePageRule($zoneId, $identifier);
	}



	public function getCacheRule($identifier)
	{
		if (!$this->cacheRules)
		{
			$this->cacheRules = $this->getCacheRules();
		}

		$rule = null;

		if ($this->cacheRules)
		{
			foreach ($this->cacheRules as $rule)
			{
				if ($rule['id'] == $identifier)
				{
					break;
				}
			}
		}

		return $rule;
	}



	public function enableCacheRule($identifier)
	{
		$rule = $this->getCacheRule($identifier);

		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updateCacheRule(
			$zoneId,
			$rule['ruleset_id'],
			$rule['id'],
			$rule['action'],
			$rule['action_parameters'],
			$rule['description'],
			true,
			$rule['expression']
		);
	}

	public function disableCacheRule($identifier)
	{
		$rule = $this->getCacheRule($identifier);

		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$api->updateCacheRule(
			$zoneId,
			$rule['ruleset_id'],
			$rule['id'],
			$rule['action'],
			$rule['action_parameters'],
			$rule['description'],
			false,
			$rule['expression']
		);
	}

	public function deleteCacheRule($rulesetId, $identifier, $hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();
		$api->deleteCacheRule($zoneId, $rulesetId, $identifier);
	}

	public function getAccessIdentityProviders($hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		return $api->getAccessIdentityProviders($accountId)['result'];
	}
	public function addAdminAccessApp($type = '', $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();

		$parsedUrl = parse_url($this->getSiteUrl());

		$url = $parsedUrl['host'] . (!empty($parsedUrl['path']) ? $parsedUrl['path'] : '');

		$emails = $this->getAdminEmails();

		if (!$emails)
		{
			$this->printableException($this->phrase('no_admins_with_email'));
		}

		$areas = $this->getAdminAreas($type);

		foreach ($areas as $urlSuffix => $phrase)
		{
			$policies = [
				[
					'decision' => 'allow',
					'exclude' => [],
					'include' => $emails,
					'name' => $this->phrase('allow'),
					'precedence' => 1,
					'require' => [],
					'uid' => ''
				]
			];

			$api->createAccessApp($accountId, $url . $urlSuffix, $phrase, $policies, '730h');
		}
	}

	public function deleteAccessApp(string $identifier, string $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		$api->deleteAccessApp($accountId, $identifier);
	}

	public function getWorkerById($workerId, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();

		if ($this->workers === null)
		{
			$this->workers = $api->getWorkers($accountId)['result'];
		}

		if ($this->workers)
		{
			foreach ($this->workers as $worker)
			{
				if ($worker['id'] == $workerId)
				{
					return $worker;
				}
			}
		}
		return [];
	}

	public function uploadWorker($workerId, $code, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		$api->uploadWorker($accountId, $workerId, $code);
	}


	public function getWorkerSubdomain($hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		$result = $api->getWorkerSubdomain($accountId);

		if ($result === false)
		{
			return false;
		}

		if (empty($result['result']['subdomain']))
		{
			return false;
		}

		return $result['result']['subdomain'];
	}

	public function workerSubdomainEnable($workerId, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		$result = $api->getWorkerEnvironmentProductionSubdomain($accountId, $workerId);

		if (empty($result['result']['enabled']))
		{
			$api->setWorkerEnvironmentProductionSubdomain($accountId, $workerId, true);
		}
	}

	public function getZoneSubscription($zoneId = null)
	{
		$api = $this->getApiClass();
		$result = $api->getZoneSubscription($zoneId ?: $this->options()->cfZoneId, false);
		return $result['result'];
	}

	public function getTurnstileSites(&$hostname = null)
	{
		$accountId = $this->getAccountId($hostname, true);
		$api = $this->getApiClass();
		$result = $api->getTurnstileSites($accountId);

		if ($result === false)
		{
			return false;
		}

		if (empty($result['result']))
		{
			return false;
		}

		return $result['result'];
	}

	public function getTurnstileWidgetByDomain($hostname = null)
	{
		$sites = $this->getTurnstileSites($hostname);
		if ($sites)
		{
			foreach ($sites as $site)
			{
				if (!empty($site['domains']) && in_array($hostname, $site['domains']))
				{
					$result = $this->getTurnstileWidgetBySitekey($site['sitekey'], $hostname);

					if ($result['result'])
					{
						return $result['result'];
					}
				}
			}
		}
		return false;
	}

	public function getTurnstileWidgetBySitekey($sitekey, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		return $api->getTurnstileWidget($accountId, $sitekey);
	}

	public function addTurnstileSite($siteName, $domain, $mode = 'managed')
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		return $api->createTurnstileWidget($accountId, $siteName, $domain, $mode);
	}

	public function getR2Buckets(&$hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();

		try
		{
			$result = $api->getR2Buckets($accountId);
		}
		catch(\Exception $e)
		{
			$error = json_decode($e->getPrevious()->response['contents'], true);
			if (isset($error['errors'][0]['code']) && $error['errors'][0]['code'] == 10042) // Cloudflare account does not have R2 enabled
			{
				return false;
			}
			throw $e;
		}

		$buckets = [];

		if (isset($result['result']['buckets']))
		{
			foreach($result['result']['buckets'] as $bucket)
			{
				$buckets[$bucket['name']] = $bucket['creation_date'];
			}
		}

		return $buckets;
	}

	public function getR2BucketInfo($bucketName = '', $hostname = null)
	{
		$api = $this->getApiClass();
		$accountId = $this->getAccountId($hostname);

		$promises = [];

		$promises[$bucketName . ':operations-classA'] = $api->getGraphQLR2Operations($accountId, $bucketName, 'classA');
		$promises[$bucketName . ':operations-classB'] = $api->getGraphQLR2Operations($accountId, $bucketName, 'classB');
		$promises[$bucketName . ':usage'] = $api->getR2Usage($accountId, $bucketName, true);
		$promises[$bucketName . ':custom-domains'] = $api->getR2CustomDomains($accountId, $bucketName, true);
		//	$promises[$bucketName . ':head'] = $api->headR2Bucket($bucketName, true);

		return $promises;
	}

	public function parseR2BucketInfo($bucketInfo)
	{
		$customDomains = [];
		if (!empty($bucketInfo['custom-domains']['result']['domains']))
		{
			foreach($bucketInfo['custom-domains']['result']['domains'] as $domain)
			{
				$customDomains[] = $domain['domain'];
			}
		}

		return [
			'operations-classA' => !empty($bucketInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) ? $bucketInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests'] : 0,
			'operations-classB' => !empty($bucketInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) ? $bucketInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests'] : 0,
			'egress-classA' => !empty($bucketInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize']) ? $bucketInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize'] : 0,
			'egress-classB' => !empty($bucketInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize']) ? $bucketInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize'] : 0,
			'usage' => empty($bucketInfo['usage']['result']) ? null : $bucketInfo['usage']['result'],
			'custom-domains' => $customDomains,
		];
	}

	public function getR2AccountInfo($hostname = null)
	{
		$api = $this->getApiClass();

		$accountId = $this->getAccountId($hostname);

		$promises = [];

		$promises['operations-classA'] = $api->getGraphQLR2Operations($accountId, null, 'classA');
		$promises['operations-classB'] = $api->getGraphQLR2Operations($accountId, null, 'classB');
		$promises['operations-usage'] = $api->getGraphQLR2Storage($accountId);
		$promises['subscription'] = $api->getZoneSubscription($this->getZoneId());

		return $promises;
	}

	public function parseR2AccountInfo($accountInfo)
	{
		$totalObjects = $totalPayloadSize = $totalMetadataSize = 0;
		$buckets = [];

		// Pretty silly we need to do this to get totals for account.  Hopefully Cloudflare adds normal account-level reporting at some point.
		// Getting the last 24 hours of entries, but ignoring all but the most recent for each bucket (buckets don't generate logs at the same times)
		if (!empty($accountInfo['operations-usage']['data']['viewer']['accounts'][0]['r2StorageAdaptiveGroups']))
		{
			foreach($accountInfo['operations-usage']['data']['viewer']['accounts'][0]['r2StorageAdaptiveGroups'] as $bucket)
			{
				$bucketName = $bucket['dimensions']['bucketName'];

				if (empty($buckets[$bucketName]))
				{
					$buckets[$bucketName] = [
						'objectCount' => $bucket['max']['objectCount'],
						'payloadSize' => $bucket['max']['payloadSize'],
						'metadataSize' => $bucket['max']['metadataSize']
					];
				}
			}

			foreach ($buckets as $bucket)
			{
				$totalObjects += $bucket['objectCount'];
				$totalPayloadSize += $bucket['payloadSize'];
				$totalMetadataSize += $bucket['metadataSize'];
			}
		}

		return [
			'operations-classA' => !empty($accountInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) ? $accountInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests'] : 0,
			'operations-classB' => !empty($accountInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) ? $accountInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests'] : 0,
			'egress-classA' => !empty($accountInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize']) ? $accountInfo['operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize'] : 0,
			'egress-classB' => !empty($accountInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize']) ? $accountInfo['operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['responseObjectSize'] : 0,
			'payload-size' => $totalPayloadSize,
			'metadata-size' => $totalMetadataSize,
			'object-count' => $totalObjects,
			'has_subscription' => !empty($accountInfo['subscription']['success']),
		];
	}

	public function addR2Bucket(string $bucketName, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();
		return $api->createR2Bucket($accountId, $bucketName);
	}

	public function addR2CustomDomain(string $bucketName, string $customDomain, $hostname = null)
	{
		$zoneId = $this->getZoneId($customDomain);

		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();

		$api->createR2CustomDomain($accountId, $bucketName, $customDomain, $zoneId, $this->zone);
		$policy = $api->getR2Policy($accountId, $bucketName);

		if (!empty($policy['result']['onlyViaCnames']))
		{
			$existingDomains = $policy['result']['onlyViaCnames'];
			if (!in_array($customDomain, $existingDomains))
			{
				$existingDomains[] = $customDomain;
			}
			$customDomain = $existingDomains;
		}

		return $api->setR2Policy($accountId, $bucketName, $customDomain, 'CnamesOnly');
	}

	public function removeR2CustomDomain(string $bucketName, string $customDomain, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$api = $this->getApiClass();

		$api->deleteR2CustomDomain($accountId, $bucketName, $customDomain);
		return $api->setR2Policy($accountId, $bucketName, '', 'CnamesOnly');
	}

	public function getR2Object($bucketName, $filename, $returnPromise = false)
	{
		$api = $this->getApiClass();
		return $api->getR2Object($bucketName, $filename, $returnPromise);
	}
	public function addR2Object($bucketName, $filename, $data, $time = 0, $returnPromise = false)
	{
		$api = $this->getApiClass();
		return $api->createR2Object($bucketName, $filename, $data, $time, $returnPromise);
	}
	public function deleteR2Object($bucketName, $filename, $returnPromise = false)
	{
		$api = $this->getApiClass();
		return $api->deleteR2Object($bucketName, $filename, $returnPromise);
	}

	// This is an Enterprise only API method
	public function traceroute($targets, $colos = [], $options = [])
	{
		$api = $this->getApiClass();
		return $api->traceroute($this->getAccountId(), $targets, $colos, $options);
	}

	public function getTurnstileAnalytics($siteKey, $date)
	{
		$api = $this->getApiClass();
		return $api->getTurnstileAnalytics($this->getAccountId(), $siteKey, '@' . ($date + 1));
	}

	public function prepareTurnstileStats($results)
	{
		$return = [
			'turnstile_challenge' => 0,
			'turnstile_solve_interactive' => 0,
			'turnstile_solve_noninteractive' => 0,
			'turnstile_unsolved' => 0,
		];

		if (!empty($results['data']['viewer']['accounts'][0]['solveEvents']) && is_array($results['data']['viewer']['accounts'][0]['solveEvents']))
		{
			$map = [
				'challenge_non_interactive_solved' => 'turnstile_solve_noninteractive',
				'challenge_issued' => 'turnstile_challenge',
				'challenge_interactive_siteverify_solved' => 'turnstile_solve_interactive',
				//	'turnstile_unsolved' => 'turnstile_unsolved' // calculated at end
			];

			foreach($results['data']['viewer']['accounts'][0]['solveEvents'] as $event)
			{
				if (!empty($map[$event['dimensions']['eventType']]))
				{
					$return[$map[$event['dimensions']['eventType']]] += $event['count'];
				}
			}

			$return['turnstile_unsolved'] = max(0, $return['turnstile_challenge'] - $return['turnstile_solve_interactive'] - $return['turnstile_solve_noninteractive']);
		}

		return $return;
	}

	public function getStats($date)
	{
		if (empty($this->option('cloudflareAuth.token')))
		{
			return [];
		}

		$siteKey = null;
		try {
			$widget = $this->getTurnstileWidgetByDomain();

			if ($widget)
			{
				$siteKey = $widget['sitekey'];
			}
		}
		catch(\Exception $e) {}

		$promises = [
			'zone_analytics' => $this->getGraphQLZoneAnalytics(1, '@' . ($date + 1))
		];

		if ($siteKey)
		{
			$promises['turnstile'] = $this->getTurnstileAnalytics($siteKey, $date);
		}

		$buckets = $this->option('cfR2Bucket');

		if (is_array($buckets))
		{
			$api = $this->getApiClass();
			$accountId = $this->getAccountId();

			$buckets = array_filter($buckets);

			if ($buckets)
			{
				$buckets = array_values($buckets);

				$promises['r2_operations-classA'] = $api->getGraphQLR2Operations($accountId, $buckets, 'classA', '@' . ($date + 1));
				$promises['r2_operations-classB'] = $api->getGraphQLR2Operations($accountId, $buckets, 'classB', '@' . ($date + 1));
			}
		}

		$promises = $this->resolvePromises($promises);

		$turnstileStats = $this->prepareTurnstileStats($siteKey ? $promises['turnstile'] : []);

		$return = [
			'cloudflare_unique_visitor' => !empty($promises['zone_analytics']['data']['viewer']['zones'][0]['totals'][0]['uniq']['uniques']) ? intval($promises['zone_analytics']['data']['viewer']['zones'][0]['totals'][0]['uniq']['uniques']) : 0,
			'cloudflare_request' => !empty($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['requests']) ? intval($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['requests']) : 0,
			'cloudflare_data_served' => !empty($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['bytes']) ? intval($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['bytes']) : 0,
			'cloudflare_data_cached' => !empty($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['cachedBytes']) ? intval($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['cachedBytes']) : 0,
			'cloudflare_threats' => !empty($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['threats']) ? intval($promises['zone_analytics']['data']['viewer']['zones'][0]['zones'][0]['sum']['threats']) : 0,
			'turnstile_challenge' => $turnstileStats['turnstile_challenge'],
			'turnstile_solve_interactive' => $turnstileStats['turnstile_solve_interactive'],
			'turnstile_solve_noninteractive' => $turnstileStats['turnstile_solve_noninteractive'],
			'turnstile_unsolved' => $turnstileStats['turnstile_unsolved'],
			'r2_operation_class_a' => !empty($promises['r2_operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) ? intval($promises['r2_operations-classA']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) : 0,
			'r2_operation_class_b' => !empty($promises['r2_operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) ? intval($promises['r2_operations-classB']['data']['viewer']['accounts'][0]['r2OperationsAdaptiveGroups'][0]['sum']['requests']) : 0,
		];

		// Shouldn't be necessary, but just in case we somehow get a negative number for anything
		foreach ($return as &$value)
		{
			$value = max(0, $value);
		}

		return $return;
	}

	public function getBackup($hostname = null)
	{
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$backup = [];

		$results = $this->processBackupResults($api->getFirewallRules($zoneId),
			['action', 'description', 'enabled', 'expression', 'id', 'ref']
		);
		if ($results !== false)
		{
			$backup['firewall_rules'] = $results;
		}

		$results = $this->processBackupResults($this->getAccessApps($hostname, false),
			['domain', 'name', 'policies', 'same_site_cookie_attribute', 'custom_deny_message', 'custom_deny_url', 'skip_interstitial', 'auto_redirect_to_identity', 'app_launcher_visible', 'service_auth_401_redirect', 'cors_headers', 'allowed_idps', 'type', 'enable_binding_cookie', 'logo_url', 'session_duration', 'http_only_cookie_attribute']
		);
		if ($results !== false)
		{
			$backup['access_apps'] = $results;
		}

		$results = $this->processBackupResults($api->getPageRules($zoneId),
			['targets', 'actions', 'priority', 'status']
		);
		if ($results !== false)
		{
			$backup['page_rules'] = $results;
		}

		$results = $this->processBackupResults($this->getZoneFirewallAccessRules($hostname),
			['mode', 'configuration', 'notes']
		);
		if ($results !== false)
		{
			$backup['firewall_access_rules'] = array_values($results);
		}

		$results = $this->processBackupResults($api->getFirewallUserAgentRules($zoneId),
			['mode', 'configuration', 'id', 'paused', 'description']
		);
		if ($results !== false)
		{
			$backup['firewall_useragent_rules'] = $results;
		}

		$results = $this->processBackupResults($api->getCacheRules($zoneId),
			['id', 'version', 'action', 'expression', 'description', 'ref', 'enabled', 'action_parameters']
		);
		if ($results !== false)
		{
			$backup['cache_rules'] = $results;
		}

		return $backup;
	}

	protected function processBackupResults($results, array $fields)
	{
		$return = [];

		if (!empty($results['success']))
		{
			if (count($results['result']) && count($results['result']))
			{
				if (!empty($results['result']['rules']))
				{
					foreach ($results['result']['rules'] as $resultKey => $result)
					{
						foreach ($result as $key => $value)
						{
							if (in_array($key, $fields))
							{
								$return[$resultKey][$key] = $value;
							}
						}
					}
				}
				else
				{
					foreach ($results['result'] as $resultKey => $result)
					{
						foreach ($result as $key => $value)
						{
							if (in_array($key, $fields))
							{
								$return[$resultKey][$key] = $value;
							}
						}
					}
				}
				return $return;
			}
			return [];
		}
		return false;
	}


	public function restoreBackup(array $backup, $hostname = null)
	{
		$accountId = $this->getAccountId($hostname);
		$zoneId = $this->getZoneId($hostname);
		$api = $this->getApiClass();

		$phase = $api->getRulesetPhase($zoneId, 'http_request_firewall_custom/entrypoint', 404);

		if ($backup['firewall_rules'])
		{
			foreach($backup['firewall_rules'] as $item)
			{
				// Using try/catch because it will throw an exception if it already exists when you first try to add the rule.
				try
				{
					if ($phase && !empty($phase['result']['id']))
					{
						$api->createFirewallRule(
							$zoneId,
							$phase['result']['id'],
							$item['expression'],
							$item['description'],
							!empty($item['ref']) ? $item['ref'] : '',
							$item['enabled'],
							$item['action']
						);
					}
					else
					{
						$api->createFirewallRule(
							$zoneId,
							'',
							$item['expression'],
							$item['description'],
							!empty($item['ref']) ? $item['ref'] : '',
							$item['enabled'],
							$item['action']
						);

						$phase = $api->getRulesetPhase($zoneId, 'http_request_firewall_custom/entrypoint', 404);
					}
				}
				catch(\Exception $e) {}
			}
		}
		if ($backup['access_apps'])
		{
			foreach($backup['access_apps'] as $item)
			{
				$api->createAccessApp(
					$accountId,
					$item['domain'],
					$item['name'],
					$item['policies'],
					$item['session_duration']
				);
			}
		}
		if ($backup['page_rules'])
		{
			foreach ($backup['page_rules'] as $item)
			{
				$api->createPageRule(
					$zoneId,
					$item['targets'],
					$item['actions'],
					$item['priority'],
					$item['status']
				);
			}
		}
		if ($backup['firewall_access_rules'])
		{
			foreach ($backup['firewall_access_rules'] as $item)
			{
				$api->createFirewallAccessRule(
					$zoneId,
					$item['configuration'],
					$item['mode'],
					$item['notes']
				);
			}
		}
		if ($backup['firewall_useragent_rules'])
		{
			foreach($backup['firewall_useragent_rules'] as $item)
			{
				$api->createFirewallUserAgentRule(
					$zoneId,
					$item['configuration'],
					$item['mode'],
					$item['description']
				);
			}
		}
		if ($backup['cache_rules'])
		{
			foreach ($backup['cache_rules'] as $item)
			{
				$api->createCacheRule(
					$zoneId,
					$item['description'],
					$item['expression'],
					$item['action_parameters'],
					$item['id'],
					$item['enabled']
				);
			}
		}
	}

	public function getPresignedUrl($method, $bucketName, $objectPath, $secondsUntilExpiration = 60, $extraQuery = [], $extraHeader = [])
	{
		return $this->getApiClass()->getPresignedUrl($method, $bucketName, $objectPath, $secondsUntilExpiration, $extraQuery, $extraHeader);
	}

}